@@ -1,5 +1,3 @@
-# !/usr/bin/env python
-# -*- coding: utf-8 -*-
 import logging
 import os
 import shutil
@@ -18,15 +16,20 @@ from alive_progress import alive_bar
 from uvicorn import Config
 
 from databutton.decorators.jobs.schedule import Scheduler
+from databutton.secrets.apiclient import SecretsApiClient
+from databutton.secrets.client import get_secrets_client
 from databutton.server.dev import DatabuttonRunner
 from databutton.utils import (
+    LoginData,
     create_databutton_cloud_project,
     create_databutton_config,
+    get_api_url,
     get_auth_token,
     get_build_logs,
     get_databutton_config,
     get_databutton_login_info,
     get_databutton_login_path,
+    new_databutton_version_exists,
 )
 from databutton.utils.deploy import (
     create_archive,
@@ -35,8 +38,7 @@ from databutton.utils.deploy import (
     upload_archive,
 )
 from databutton.utils.uvicorn_in_mem import get_threaded_uvicorn
-
-from .__init__ import __version__
+from databutton.version import __version__
 
 logger = logging.getLogger("databutton.cli")
 
@@ -49,12 +51,23 @@ LOGGING_LEVELS = {
 }
 
 
+def require_databutton_login_info() -> LoginData:
+    login_info = get_databutton_login_info()
+    if login_info is None:
+        click.secho("❌ You're not authenticated to Databutton, yet.")
+        click.secho(
+            f"Type {click.style('databutton login', fg='cyan')} to login or create a user."
+        )
+        exit(1)
+    return login_info
+
+
 @click.group()
 @click.option("--verbose", "-v", count=True, help="Enable verbose logging")
 @click.pass_context
 def cli(ctx, verbose: int):
-    ctx.ensure_object(dict)
     """Run databutton."""
+    ctx.ensure_object(dict)
     # Use the verbosity count to determine the logging level...
     if verbose > 0:
         logger.setLevel(
@@ -130,22 +143,26 @@ def create(
     login_info = get_databutton_login_info()
     if not login_info:
         click.echo()
-        click.secho("It doesn't seem like you have a user with Databutton.")
         click.secho(
-            f"In order to use {click.style('databutton deploy', fg='cyan')} and "
-            + f"{click.style('db.storage.dataframes', fg='cyan')}, "
-            "you need a user and a project in Databutton."
-        )
-        should_login = (
-            False
-            if non_interactive
-            else click.confirm("\nWould you like to create a user now?", default=True)
+            "It doesn't seem like you have a user with Databutton, at least you're not logged in."
         )
+        if non_interactive:
+            click.echo(
+                "Can't log interactively in non-interactive mode."
+                + "\nRun databutton login first or run databutton create without --non-interactive"
+            )
+            exit(1)
+        click.secho("Let's log you in or create a user...")
         click.echo()
-        if should_login:
-            ctx.invoke(login)
-            login_info = get_databutton_login_info()
+        ctx.invoke(login)
+        login_info = get_databutton_login_info()
+    if not login_info:
+        # This might still not be true...
+        click.echo("Could not log in successfully, please try again..")
+        exit(1)
+    project_id = create_databutton_cloud_project(project_name)
 
+    click.secho("Downloading template...")
     res = requests.get(
         f"https://storage.googleapis.com/databutton-app-templates/{template}.tar.gz",
         stream=True,
@@ -155,84 +172,95 @@ def create(
             f"❌ Failed to get template {template}, are you sure you typed an existing template?"
         )
         exit(1)
-    if res.ok:
-        with tempfile.TemporaryFile() as tmpfile:
-            tmpfile.write(res.raw.read())
-            tmpfile.seek(0)
-            with tarfile.open(fileobj=tmpfile) as t:
-                t.extractall(project_directory)
+    with tempfile.TemporaryFile() as tmpfile:
+        tmpfile.write(res.raw.read())
+        tmpfile.seek(0)
+        with tarfile.open(fileobj=tmpfile) as t:
+            t.extractall(project_directory)
+    create_databutton_config(
+        name=project_name, uid=project_id, project_directory=project_directory
+    )
+    click.echo()
+    click.echo(
+        f"Success! Created project {click.style(project_name, fg='cyan')} in {project_directory}"
+    )
+    click.echo("Inside that directory, you can run several commands:")
+    click.echo()
 
-        if login_info:
-            should_create_project = (
-                False
-                if non_interactive
-                else click.confirm(
-                    "\nDo you want to also create a project on databutton.com?\nYou can always do databutton init later.",
-                    default=True,
-                )
-            )
-            if should_create_project:
-                project_id = create_databutton_cloud_project(project_name)
-                create_databutton_config(
-                    name=project_name,
-                    uid=project_id,
-                    project_directory=project_directory,
-                )
-            else:
-                create_databutton_config(
-                    name=project_name,
-                    uid=str(uuid4()),
-                    project_directory=project_directory,
-                )
-        else:
-            create_databutton_config(
-                name=project_name, uid=str(uuid4()), project_directory=project_directory
-            )
-        click.echo()
-        click.echo(
-            f"Success! Created project {click.style(project_name, fg='cyan')} in {project_directory}"
-        )
-        click.echo("Inside that directory, you can run several commands:")
+    commands = [
+        ("databutton start", "Starts the development server."),
+        (
+            "databutton deploy",
+            "Deploy your project to databutton.com.\n    "
+            + "Note that in order to deploy you need a user and a project in Databutton.",
+        ),
+        (
+            "databutton build",
+            "Bundles the project and generates the necessary files for production.",
+        ),
+    ]
+
+    for cmd, desc in commands:
+        click.echo(f"  {click.style(cmd, fg='cyan')}")
+        click.echo(f"    {click.style(desc)}")
         click.echo()
 
-        commands = [
-            ("databutton start", "Starts the development server."),
-            (
-                "databutton deploy",
-                "Deploy your project to databutton.com.\n    "
-                + "Note that in order to deploy you need a user and a project in Databutton.",
-            ),
-            (
-                "databutton build",
-                "Bundles the project and generates the necessary files for production.",
-            ),
-        ]
+    click.echo("We suggest you begin by typing:\n")
 
-        for cmd, desc in commands:
-            click.echo(f"  {click.style(cmd, fg='cyan')}")
-            click.echo(f"    {click.style(desc)}")
-            click.echo()
+    cd_dir = (
+        project_directory.relative_to(Path.cwd())
+        if project_directory.is_relative_to(Path.cwd())
+        else project_directory
+    )
 
-        click.echo("We suggest you begin by typing:\n")
+    click.echo(f"  {click.style('cd', fg='cyan')} {cd_dir}")
+    click.secho("  pip install -r requirements.txt", fg="cyan")
+    click.secho("  databutton start", fg="cyan")
+    click.secho("  databutton deploy", fg="cyan")
 
-        click.echo(f"  {click.style('cd', fg='cyan')} {project_directory}")
-        click.secho("  pip install -r requirements.txt", fg="cyan")
-        click.secho("  databutton start", fg="cyan")
-        click.secho("  databutton deploy", fg="cyan")
+    click.echo()
+    click.secho(
+        f"You can always see the available commands by typing {click.style('databutton --help', fg='cyan')}."
+        + f"\nDocumentation is available by typing {click.style('databutton docs', fg='cyan')}."
+    )
 
-        click.echo()
-        click.secho(
-            f"You can always see the available commands by typing {click.style('databutton --help', fg='cyan')}."
-            + f"\nDocumentation is available by typing {click.style('databutton docs', fg='cyan')}."
-        )
-
-        click.echo("\nHappy building!\n")
+    click.echo("\nHappy building!\n")
 
 
 @cli.command()
 @click.option("--debug", default=False, type=bool, show_default=True, is_flag=True)
 def start(debug=False):
     """Run the Databutton development server"""
+    new_version = new_databutton_version_exists()
+    if new_version:
+        click.echo()
+        click.echo("There is a new version of Databutton available")
+        click.echo()
+        if Path("pyproject.toml").exists():
+            # Poetry project
+            click.echo(
+                f'write {click.style(f"poetry add databutton@{new_version}", fg="yellow")} to install the new version.'
+            )
+        else:
+            click.echo(
+                f'write {click.style(f"pip install databutton=={new_version}", fg="yellow")} to install the new version'
+            )
+        click.echo()
+        click.echo()
+
+    # Make sure a config exists
+    try:
+        get_databutton_config()
+    except FileNotFoundError:
+        click.secho("Could not find a databutton.json file.")
+        click.secho(
+            f"Are you sure you ran {click.style('databutton start', fg='cyan')} in the right directory?"
+        )
+        exit(1)
+    except Exception as e:
+        print(e)
+        exit(1)
+
     runner = DatabuttonRunner()
 
     try:
@@ -275,16 +303,10 @@ def login():
 def deploy(ctx: click.Context):
     """Deploy your project to Databutton"""
 
-    login_info = get_databutton_login_info()
-    if login_info is None:
-        click.secho("❌ You're not authenticated to Databutton, yet.")
-        click.secho(
-            f"Type {click.style('databutton login', fg='cyan')} to login or create a user."
-        )
-        exit(1)
+    _ = require_databutton_login_info()
 
     try:
-        config = get_databutton_config(retries=2)
+        config = get_databutton_config()
     except FileNotFoundError:
         click.secho(
             "❌ Can't find a Databutton config file. Are you in the correct folder?"
@@ -299,7 +321,7 @@ def deploy(ctx: click.Context):
 
     click.echo()
     click.echo(click.style(f"=== Deploying to {config.name}", fg="green"))
-    ctx.invoke(build)
+    # ctx.invoke(build)
     click.echo(click.style("i packaging components...", fg="cyan"))
 
     with tempfile.NamedTemporaryFile() as tmpfile:
@@ -390,12 +412,9 @@ def serve():
 @click.option("--name", help="Name of the project")
 def init(name: str):
     """Creates a new project in Databutton and writes to databutton.json"""
-    login_info = get_databutton_login_info()
-    if login_info is None:
-        click.secho(
-            f"You're not logged in. Log in with {click.style('databutton login', fg='cyan')} first."
-        )
-        exit(1)
+
+    _ = require_databutton_login_info()
+
     if os.path.exists("databutton.json"):
         click.secho("There is already a databutton.json file in this directory")
         project_config = get_databutton_config()
@@ -457,27 +476,222 @@ def logout():
 
 
 @cli.command()
-def whoami():
+@click.option("-t", "--token", help="Print an id token", is_flag=True, default=False)
+@click.option(
+    "-r",
+    "--refresh-token",
+    default=False,
+    is_flag=True,
+    help="Print the refresh token."
+    + " Set DATABUTTON_TOKEN to this if you want to run databutton in ci or codespaces.",
+)
+def whoami(token: bool, refresh_token: bool):
     """Shows the logged in user"""
     import jwt
 
     try:
-        token = get_auth_token()
+        id_token = get_auth_token()
     except Exception:
         click.secho("No user found.")
         click.secho(f"Log in first with {click.style('databutton login', fg='cyan')}.")
         exit(1)
-    decoded = jwt.decode(token, options={"verify_signature": False})
+    decoded = jwt.decode(id_token, options={"verify_signature": False})
     click.secho("Found logged in user")
     keys_to_print = ["name", "email", "user_id"]
     for k in keys_to_print:
         click.secho(f"  {k}: {click.style(decoded[k], fg='cyan')}")
+    if token:
+        click.secho(f"  token: {click.style(id_token, fg='cyan')}")
+    if refresh_token:
+        click.secho(
+            f"  refreshToken: {click.style(get_databutton_login_info().refreshToken, fg='cyan')}"
+        )
     click.echo()
 
 
+@cli.command()
+def print_token():
+    click.echo(get_auth_token())
+
+
 @cli.command()
 def schedule():
     """Starts and runs the scheduler"""
     click.secho("=== Starting the databutton scheduler", fg="cyan")
 
     Scheduler.create()
+
+
+@cli.command()
+@click.argument("project_id")
+@click.argument("project_dir", required=False)
+@click.pass_context
+def download(ctx: click.Context, project_id: str, project_dir: str):
+    """
+    Downloads a project locally from PROJECT_ID and into PROJECT_DIR.
+    This will download the last deployed code
+    """
+    project_dir: Path = Path(project_dir) if project_dir is not None else Path.cwd()
+    if project_dir.exists():
+        if len(list(project_dir.iterdir())) > 0:
+            click.secho("")
+            click.secho("❌ Can not download into a folder that's not empty.")
+            click.secho(
+                "Please select a folder that's empty (or one that doesn't exist) :-)"
+            )
+            exit(1)
+
+    # Double check that the user is logged in
+    try:
+        get_auth_token()
+    except Exception:
+        click.echo("It doesn't seem like you're logged in, but we can fix that!")
+        if click.confirm("Do you want to log in right now?"):
+            ctx.invoke(login)
+        else:
+            click.echo("You need to be logged in to download this project")
+            click.secho(f"Type {click.style('databutton login', fg='cyan')} to login")
+            exit(1)
+
+    click.secho("i downloading project...", fg="cyan")
+    click.secho(f"i id: {project_id}", fg="cyan")
+    click.secho(f"i directory: {str(project_dir)}", fg="cyan")
+    download_url = f"{get_api_url(project_id)}/projects/download"
+    response = requests.get(
+        download_url,
+        headers={"Authorization": f"Bearer {get_auth_token()}"},
+        stream=True,
+    )
+    # Ensure that there's a zipped response
+    # This can yield a 307 to login if the user does not have access
+    # We can catch that by making sure that the response needs to be a tarball
+    if (
+        not response.ok
+        or response.headers.get("content-type") != "application/tar+gzip"
+    ):
+        # This would be the error message from the upstream /download API
+        if "next.databutton.com/login" in response.url:
+            # This is what an unauthenticated api requests looks like :'(
+            # We get redirected
+            click.secho("Are you sure you're logged in as the correct user?")
+            click.secho(
+                f"Type {click.style('databutton whoami')} to see who you're logged in as."
+            )
+            exit(1)
+        elif "detail" in response.text:
+            click.secho("This is the response from the server:")
+            click.secho("  " + response.json()["detail"])
+        click.echo()
+        click.secho(
+            f"❌ Failed to get project with id {project_id}, are you sure it's the correct id?"
+        )
+        click.secho(
+            "If you're sure it's the right id, the following url should work and have some components:"
+        )
+        click.secho(f"https://next.databutton.com/projects/{project_id}", fg="cyan")
+        exit(1)
+    with tempfile.TemporaryFile() as tmpfile:
+        tmpfile.write(response.raw.read())
+        tmpfile.seek(0)
+        with tarfile.open(fileobj=tmpfile) as t:
+            t.extractall(project_dir)
+    click.secho("Finished!", fg="green")
+    click.echo()
+    click.secho("Finished downloading project, we recommmend that you now do")
+    click.echo()
+    cd_dir = (
+        project_dir.relative_to(Path.cwd())
+        if project_dir.is_relative_to(Path.cwd())
+        else project_dir
+    )
+    click.secho(f"  cd {cd_dir}", fg="cyan")
+    click.secho("  pip install -r requirements.txt", fg="cyan")
+    click.secho("  databutton start", fg="cyan")
+    click.secho("  and when you're happy with your changes,")
+    click.secho("  databutton deploy", fg="cyan")
+    click.echo()
+
+
+@cli.group("secrets")
+@click.pass_context
+def secrets_group(ctx: click.Context):
+    """Manage project secrets"""
+    # Shared prerequisites for all secrets commands
+    # _ = require_databutton_login_info()
+    pass
+
+
+@secrets_group.command("add")
+@click.pass_context
+@click.argument("name", required=True)
+@click.argument("data-file", type=click.File("r"))
+def secrets_add(ctx: click.Context, name: str, data_file):
+    """Add secret to project.
+
+    NB! Input is name of secret and the name of a _file_
+    containing the secret value.
+
+    The secret value is passed to the databutton api and
+    stored encrypted in the cloud.
+    """
+    secrets_client: SecretsApiClient = get_secrets_client()
+
+    # Read raw bytes from file
+    max_len = 64 * 1024
+    value: str = data_file.read(max_len + 1)
+    if len(value) > max_len:
+        click.secho(f"Secret value is too long (max={max_len})")
+        exit(1)
+
+    # Remove trailing newline which is usually not part of a secret.
+    # Note: If this becomes a problem, add a parameter
+    #       to make it opt-out but keep it as default.
+    value = value.rstrip("\r\n")
+
+    # Store secret value
+    success = secrets_client.add(name, value)
+    if success:
+        click.echo("Successfully added secret.")
+    else:
+        click.echo("Failed to add secret!")
+
+
+@secrets_group.command("delete")
+@click.pass_context
+@click.argument("name", required=True)
+def secrets_delete(ctx: click.Context, name: str):
+    """Delete named secret from project"""
+    secrets_client: SecretsApiClient = get_secrets_client()
+
+    config = get_databutton_config()
+    confirmed = click.confirm(
+        f'This will delete the secret "{name}" from project {config.uid}, are you sure?'
+    )
+    if not confirmed:
+        click.echo("Aborted.")
+
+    success = secrets_client.delete(name)
+    if success:
+        click.echo("Successfully deleted secret.")
+    else:
+        click.echo("Failed to delete secret!")
+
+
+@secrets_group.command("get")
+@click.argument("name", required=True)
+@click.pass_context
+def secrets_get(ctx: click.Context, name: str):
+    """Show secret value."""
+    secrets_client: SecretsApiClient = get_secrets_client()
+    value = secrets_client.get(name)
+    click.echo(value)
+
+
+@secrets_group.command("list")
+@click.pass_context
+def secrets_list(ctx: click.Context):
+    """List names of secrets in project"""
+    secrets_client: SecretsApiClient = get_secrets_client()
+    secrets = secrets_client.list()
+    for s in secrets:
+        click.echo(s)
